# The set of tuples in question in these programs is the set of tuples
# of length 25, in which each number 1..5 occurs exactly 5 times.

# In the following variable we collect the list of S-minimal and U-minimal
# tuples (see below for a definition):
l := [];

# The next three variables are globals used in the recursion:
current := "                         ";
available := [5,5,5,5,5];  # in the beginning 5 of each are allowed
off := INT_CHAR('0');

AllSandUminimals := function(depth,min,used)
  # This function enumerates the list of all S-minimal and U-minimal tuples.
  # It is first called with depth=1, min=1 and used=0.
  # This function calls itself recursively.
  # depth: 1..25 next entry to choose
  # min: minimal value of next entry (to ensure U-minimality)
  # used: biggest already used number (to ensure S-minimality)
  local sum,i,newused;
  if depth > 25 then
    Add(l,ShallowCopy(current));
    return 1;   # number of solutions in the current subtree
  fi;
  sum := 0;
  for i in [min..5] do
    if available[i] > 0 and used >= i-1 then
      # we put an i on position depth and recurse down:
      if used < i then
        newused := i;
      else
        newused := used;
      fi;
      if depth < 10 then
        Print(i,"\c");
      fi;
      current[depth] := CHAR_INT(i+off);
      available[i] := available[i]-1;
      if depth mod 5 = 0 then
        # we begin a new cluster
	sum := sum + AllSandUminimals(depth+1,1,newused);
      else
        sum := sum + AllSandUminimals(depth+1,i,newused);
      fi;
      available[i] := available[i]+1;
      current[depth] := ' ';
      if depth < 10 then
        Print("\b \b");
      fi;
    fi;
  od;
  return sum;
end;

# The following is a string to copy from for creation of new tuples:
ergst := "                         ";

Unormalize := function(v)
  # A tuple of length 25 is called "U-minimal", if the entries 1..5, 6..10,
  # 11..15, 16..20, and 21..25 are in ascending order respectively.
  # This function calculates the U-minimalization, that is the U-minimal
  # tuples in the same U = S_5 x S_5 x S_5 x S_5 x S_5 orbit.
  # Returns a new tuple.
  local i,j,k,l,res,c,erg;
  res := 0*[1..25];
  erg := ShallowCopy(ergst);
  for i in [0..4] do
    for j in [1..5] do
      k := i*5+INT_CHAR(v[i*5+j])-off;
      res[k] := res[k] + 1;
    od;
    l := 1;
    for j in [1..5] do
      c := CHAR_INT(j+off);
      for k in [1..res[i*5+j]] do
        erg[i*5+l] := c;
        l := l + 1;
      od;
    od;
  od;
  return erg;
end;

Snormalize := function(v)
  # A tuple is called "S-minimal", if for all 1 <= i < j <= 5 the first
  # occurence of i is before the first occurence of j. This can always
  # be achieved by renaming the entries (operation of S_5 from the left).
  # Returns a new tuple.
  local a,         # number found at position i
        found,     # highest number found so far (in the result)
        i,         # counter
        res,       # result
        tab;       # helper tab, mapping old numbers to new ones.
  found := 0;
  tab := 0*[1..5];
  res := ShallowCopy(ergst);
  for i in [1..25] do
    a := INT_CHAR(v[i])-off;
    if tab[a] = 0 then
      found := found + 1;
      tab[a] := found;
    fi;
    res[i] := CHAR_INT(tab[a]+off);
  od;
  return res;
end;

nr := AllSandUminimals(1,1,0);  # the list is in the global variable l now
# The following global variables are used to compute the equivalence classes
# on l. We start with the finest class distribution and unite classes once
# we find that some tuples in one class lies in the same orbit as another
# tuples in the other class.
trupps := List([1..Length(l)],i->[i]);
truppof := List([1..Length(l)],i->i);
nrtrupps := Length(l);

fusionieretrupps := function(i,j)
  local k,oldtruppofj;
  if truppof[i] <> truppof[j] then
    oldtruppofj := truppof[j];
    for k in trupps[truppof[j]] do
      truppof[k] := truppof[i];
    od;
    Append(trupps[truppof[i]],trupps[oldtruppofj]);
    trupps[oldtruppofj] := [];
    nrtrupps := nrtrupps-1;
  fi;
end; 

RestoreStateRandom(
[ 35, [ 264044297, 183552375, 186557741, 46008580, 91966775, 40228334, 
      111587641, 262452744, 44854923, 39845682, 206581571, 92903376, 
      231314873, 66408590, 258714463, 184978940, 159620597, 123129748, 
      68026743, 233471294, 232591449, 24005304, 5708683, 167299138, 
      197744113, 157639412, 157761179, 88975279, 246513147, 79194103, 
      233353667, 128956703, 196460747, 115256455, 135696962, 138731753, 
      64831979, 175307525, 190299479, 107235201, 53348259, 103934810, 
      207032077, 145748420, 124215055, 144100830, 6088801, 180843272, 
      27153689, 259704742, 66996111, 23305316, 136497461, 203183570, 
      153356235 ] ]
);

# The wreath product S5 wr S5:
h := Group( (1,2,3,4,5),(1,2),
   (1,6,11,16,21)(2,7,12,17,22)(3,8,13,18,23)(4,9,14,19,24)(5,10,15,20,25),
   (1,6)(2,7)(3,8)(4,9)(5,10) );
# Choose 3 random elements:
els := [];
for i in [1..3] do
  Add(els,ListPerm(Random(h)));
od;

tryhard := function()
  local i,j,k,p,v;
  for j in [Length(l),Length(l)-1..1] do
    for k in els do
      v := Unormalize(Snormalize(l[j]{k}));
      p := POSITION_SORTED_LIST(l,v);
      fusionieretrupps(j,p);
    od;
    if j mod 10000 = 0 then
      Print("j:",j," Number of trupps:",nrtrupps,"\n");
    fi;
  od;
  Print("Number of trupps:",nrtrupps,"\n");
end;

fusmapSnwrSk := function(n,k,ctSnwrSk)
  # This calculates a fusion map from the wreath product Sn wr Sk into
  # the symmetric group S_{nk} using combinatorics. 
  local fus,gt,i,p,partsn,partsnk,parttuples,pt,t;
  parttuples := CharacterParameters(ctSnwrSk);
  partsn := Partitions(n);
  partsnk := Partitions(n*k);
  gt := function(a,b) return a > b; end;

  fus := [];
  for pt in parttuples do
    p := [];
    for i in [1..Length(pt)] do
      for t in pt[i] do
        Append(p,t * partsn[i]);
      od;
    od;
    Sort(p,gt);
    Add(fus,Position(partsnk,p));
  od;
  return fus;
end;

tryhard();

tab := Set(truppof);
result := List(truppof,i->Position(tab,i));

reps := List([1..1856],i->Position(result,i));

perms := List(reps,i->ListPerm(SortingPerm(l[i])^-1));

for p in perms do
  while Length(p) < 25 do
    Add(p,Length(p)+1);
  od;
od;

WriteThings := function()
    local s,p,st,i;
    s := OutputTextFile("perms.txt",false);
    for p in perms do
        st := "";
        for i in [1..25] do
            Append(st," ");
            Append(st,String(p[i]-1));
        od;
        WriteLine(s,st);
    od;
    CloseStream(s);

    s := OutputTextFile("fus.txt",false);
    for i in result do
        WriteLine(s,String(i));
    od;
    CloseStream(s);
end;

